import pytest

from dbt.tests.util import run_dbt

from dbt.tests.adapter.incremental.test_incremental_on_schema_change import (
    BaseIncrementalOnSchemaChangeSetup,
)


class IncrementalOnSchemaChangeIgnoreFail(BaseIncrementalOnSchemaChangeSetup):
    def test_run_incremental_ignore(self, project):
        select = "model_a incremental_ignore incremental_ignore_target"
        compare_source = "incremental_ignore"
        compare_target = "incremental_ignore_target"
        self.run_twice_and_assert(select, compare_source, compare_target, project)

    def test_run_incremental_fail_on_schema_change(self, project):
        select = "model_a incremental_fail"
        run_dbt(["run", "--models", select, "--full-refresh"])
        results_two = run_dbt(["run", "--models", select], expect_pass=False)
        assert "Compilation Error" in results_two[1].message


@pytest.mark.skip_profile("databricks_sql_endpoint", "spark_http_odbc")
class TestAppendOnSchemaChange(IncrementalOnSchemaChangeIgnoreFail):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "models": {
                "+incremental_strategy": "append",
            }
        }


@pytest.mark.skip_profile("databricks_sql_endpoint", "spark_session", "spark_http_odbc")
class TestInsertOverwriteOnSchemaChange(IncrementalOnSchemaChangeIgnoreFail):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "models": {
                "+file_format": "parquet",
                "+partition_by": "id",
                "+incremental_strategy": "insert_overwrite",
            }
        }


@pytest.mark.skip_profile("apache_spark", "spark_session")
class TestDeltaOnSchemaChange(BaseIncrementalOnSchemaChangeSetup):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "models": {
                "+file_format": "delta",
                "+incremental_strategy": "merge",
                "+unique_key": "id",
            }
        }

    def run_incremental_sync_all_columns(self, project):
        select = "model_a incremental_sync_all_columns incremental_sync_all_columns_target"
        run_dbt(["run", "--models", select, "--full-refresh"])
        # Delta Lake doesn"t support removing columns -- show a nice compilation error
        results = run_dbt(["run", "--models", select], expect_pass=False)
        assert "Compilation Error" in results[1].message

    def run_incremental_sync_remove_only(self, project):
        select = "model_a incremental_sync_remove_only incremental_sync_remove_only_target"
        run_dbt(["run", "--models", select, "--full-refresh"])
        # Delta Lake doesn"t support removing columns -- show a nice compilation error
        results = run_dbt(["run", "--models", select], expect_pass=False)
        assert "Compilation Error" in results[1].message

    def test_run_incremental_append_new_columns(self, project):
        # only adding new columns in supported
        self.run_incremental_append_new_columns(project)
        # handling columns that have been removed doesn"t work on Delta Lake today
        # self.run_incremental_append_new_columns_remove_one(project)

    def test_run_incremental_sync_all_columns(self, project):
        self.run_incremental_sync_all_columns(project)
        self.run_incremental_sync_remove_only(project)


# Test fixtures for special character column names
_MODEL_A_SPECIAL_CHARS = """
{{
    config(materialized='table')
}}

with source_data as (

    select 1 as id, 'bbb' as `select`, 111 as `field-with-dash`
    union all select 2 as id, 'ddd' as `select`, 222 as `field-with-dash`
    union all select 3 as id, 'fff' as `select`, 333 as `field-with-dash`
    union all select 4 as id, 'hhh' as `select`, 444 as `field-with-dash`
    union all select 5 as id, 'jjj' as `select`, 555 as `field-with-dash`
    union all select 6 as id, 'lll' as `select`, 666 as `field-with-dash`

)

select id
       ,`select`
       ,`field-with-dash`

from source_data
"""

_MODEL_INCREMENTAL_APPEND_NEW_SPECIAL_CHARS = """
{{
    config(
        materialized='incremental',
        unique_key='id',
        on_schema_change='append_new_columns'
    )
}}

WITH source_data AS (SELECT * FROM {{ ref('model_a_special_chars') }} )

{% if is_incremental() %}

SELECT id,
       `select`,
       `field-with-dash`
FROM source_data WHERE id NOT IN (SELECT id from {{ this }} )

{% else %}

SELECT id,
       `select`
FROM source_data where id <= 3

{% endif %}
"""

_MODEL_INCREMENTAL_APPEND_NEW_SPECIAL_CHARS_TARGET = """
{{
    config(materialized='table')
}}

with source_data as (

    select * from {{ ref('model_a_special_chars') }}

)

select id
       ,`select`
       ,CASE WHEN id <= 3 THEN NULL ELSE `field-with-dash` END AS `field-with-dash`

from source_data
"""


@pytest.mark.skip_profile("databricks_sql_endpoint", "spark_http_odbc")
class TestAppendOnSchemaChangeSpecialChars(BaseIncrementalOnSchemaChangeSetup):
    """Test incremental models with special character column names using append strategy

    Tests column quoting with SQL keywords and dashes in column names.
    Note: Spaces and dots in column names are not supported by some databases.
    """

    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "models": {
                "+incremental_strategy": "append",
            }
        }

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "model_a_special_chars.sql": _MODEL_A_SPECIAL_CHARS,
            "incremental_append_new_special_chars.sql": _MODEL_INCREMENTAL_APPEND_NEW_SPECIAL_CHARS,
            "incremental_append_new_special_chars_target.sql": _MODEL_INCREMENTAL_APPEND_NEW_SPECIAL_CHARS_TARGET,
        }

    def test_incremental_append_new_columns_with_special_characters(self, project):
        """Test that incremental models work correctly with special character column names when adding new columns"""

        # First run - creates initial table
        run_dbt(
            [
                "run",
                "--models",
                "model_a_special_chars incremental_append_new_special_chars incremental_append_new_special_chars_target",
            ]
        )

        # Second run - should append new columns with special characters
        run_dbt(
            [
                "run",
                "--models",
                "model_a_special_chars incremental_append_new_special_chars incremental_append_new_special_chars_target",
            ]
        )

        # Verify results match expected output
        from dbt.tests.util import check_relations_equal

        check_relations_equal(
            project.adapter,
            [
                "incremental_append_new_special_chars",
                "incremental_append_new_special_chars_target",
            ],
        )



import pytest

from dbt.tests.adapter.incremental.test_incremental_merge_exclude_columns import (
    BaseMergeExcludeColumns,
)


@pytest.mark.skip_profile("spark_session", "apache_spark")
class TestMergeExcludeColumns(BaseMergeExcludeColumns):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {"models": {"+file_format": "delta"}}



import pytest

from dbt.tests.adapter.unit_testing.test_types import BaseUnitTestingTypes
from dbt.tests.adapter.unit_testing.test_case_insensitivity import BaseUnitTestCaseInsensivity
from dbt.tests.adapter.unit_testing.test_invalid_input import BaseUnitTestInvalidInput


class TestSparkUnitTestingTypes(BaseUnitTestingTypes):
    @pytest.fixture
    def data_types(self):
        # sql_value, yaml_value
        return [
            ["1", "1"],
            ["2.0", "2.0"],
            ["'12345'", "12345"],
            ["'string'", "string"],
            ["true", "true"],
            ["date '2011-11-11'", "2011-11-11"],
            ["timestamp '2013-11-03 00:00:00-0'", "2013-11-03 00:00:00-0"],
            ["array(1, 2, 3)", "'array(1, 2, 3)'"],
            [
                "map('10', 't', '15', 'f', '20', NULL)",
                """'map("10", "t", "15", "f", "20", NULL)'""",
            ],
            ['named_struct("a", 1, "b", 2, "c", 3)', """'named_struct("a", 1, "b", 2, "c", 3)'"""],
        ]


class TestSparkUnitTestCaseInsensitivity(BaseUnitTestCaseInsensivity):
    pass


class TestSparkUnitTestInvalidInput(BaseUnitTestInvalidInput):
    pass



import pytest
from dbt.tests.adapter.utils.test_timestamps import BaseCurrentTimestamps


class TestCurrentTimestampSpark(BaseCurrentTimestamps):
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "get_current_timestamp.sql": "select {{ current_timestamp() }} as current_timestamp"
        }

    @pytest.fixture(scope="class")
    def expected_schema(self):
        return {"current_timestamp": "timestamp"}

    @pytest.fixture(scope="class")
    def expected_sql(self):
        return """select current_timestamp() as current_timestamp"""



_MODELS__MY_FUN_DOCS = """
{% docs my_fun_doc %}
name Column description "with double quotes"
and with 'single  quotes' as welll as other;
'''abc123'''
reserved -- characters
--
/* comment */
Some $lbl$ labeled $lbl$ and $$ unlabeled $$ dollar-quoting

{% enddocs %}
"""

_MODELS__INCREMENTAL_DELTA = """
{{ config(materialized='incremental', file_format='delta') }}
select 1 as id, 'Joe' as name
"""

_MODELS__TABLE_DELTA_MODEL = """
{{ config(materialized='table', file_format='delta') }}
select 1 as id, 'Joe' as name
"""

_MODELS__VIEW_DELTA_MODEL = """
{{ config(materialized='view') }}
select id, count(*) as count from {{ ref('table_delta_model') }} group by id
"""

_MODELS__TABLE_DELTA_MODEL_MISSING_COLUMN = """
{{ config(materialized='table', file_format='delta') }}
select 1 as id, 'Joe' as different_name
"""
_VIEW_PROPERTIES_MODELS = """
version: 2

models:
  - name: view_delta_model
    description: |
      View model description "with double quotes"
      and with 'single  quotes' as welll as other;
      '''abc123'''
      reserved -- characters
      --
      /* comment */
      Some $lbl$ labeled $lbl$ and $$ unlabeled $$ dollar-quoting
    columns:
      - name: id
        description: |
          id Column description "with double quotes"
          and with 'single  quotes' as welll as other;
          '''abc123'''
          reserved -- characters
          --
          /* comment */
          Some $lbl$ labeled $lbl$ and $$ unlabeled $$ dollar-quoting
"""
_PROPERTIES__MODELS = """
version: 2

models:
  - name: table_delta_model
    description: |
      Table model description "with double quotes"
      and with 'single  quotes' as welll as other;
      '''abc123'''
      reserved -- characters
      --
      /* comment */
      Some $lbl$ labeled $lbl$ and $$ unlabeled $$ dollar-quoting
    columns:
      - name: id
        description: |
          id Column description "with double quotes"
          and with 'single  quotes' as welll as other;
          '''abc123'''
          reserved -- characters
          --
          /* comment */
          Some $lbl$ labeled $lbl$ and $$ unlabeled $$ dollar-quoting
      - name: name
        description: |
          Some stuff here and then a call to
          {{ doc('my_fun_doc')}}

  - name: incremental_delta_model
    description: |
      Incremental model description "with double quotes"
      and with 'single  quotes' as welll as other;
      '''abc123'''
      reserved -- characters
      --
      /* comment */
      Some $lbl$ labeled $lbl$ and $$ unlabeled $$ dollar-quoting
    columns:
      - name: id
        description: |
          id Column description "with double quotes"
          and with 'single  quotes' as welll as other;
          '''abc123'''
          reserved -- characters
          --
          /* comment */
          Some $lbl$ labeled $lbl$ and $$ unlabeled $$ dollar-quoting
      - name: name
        description: |
          Some stuff here and then a call to
          {{ doc('my_fun_doc')}}
"""

_PROPERTIES__SEEDS = """
version: 2

seeds:
  - name: seed
    description: |
      Seed model description "with double quotes"
      and with 'single  quotes' as welll as other;
      '''abc123'''
      reserved -- characters
      --
      /* comment */
      Some $lbl$ labeled $lbl$ and $$ unlabeled $$ dollar-quoting
    columns:
      - name: id
        description: |
          id Column description "with double quotes"
          and with 'single  quotes' as welll as other;
          '''abc123'''
          reserved -- characters
          --
          /* comment */
          Some $lbl$ labeled $lbl$ and $$ unlabeled $$ dollar-quoting
      - name: name
        description: |
          Some stuff here and then a call to
          {{ doc('my_fun_doc')}}
"""

_SEEDS__BASIC = """id,name
1,Alice
2,Bob
"""



"""Unit tests for polling behavior in PyhiveConnectionWrapper."""

import time
import unittest
from unittest import mock

from dbt_common.exceptions import DbtRuntimeError, DbtDatabaseError
from dbt.adapters.spark.connections import PyhiveConnectionWrapper

try:
    from TCLIService.ttypes import TOperationState as ThriftState
except ImportError:
    ThriftState = None


class TestPyhivePolling(unittest.TestCase):
    """Test polling behavior fixes for long-running queries."""

    def setUp(self):
        """Set up mock cursor and connection."""
        self.mock_connection = mock.MagicMock()
        self.mock_cursor = mock.MagicMock()
        self.mock_connection.cursor.return_value = self.mock_cursor

    @mock.patch("dbt.adapters.spark.connections.time.sleep")
    def test_polling_sleeps_between_polls(self, mock_sleep):
        """Verify that polling actually sleeps between poll attempts."""
        wrapper = PyhiveConnectionWrapper(self.mock_connection, poll_interval=5)
        wrapper._cursor = self.mock_cursor

        # Mock poll responses: pending -> pending -> success
        poll_responses = [
            mock.MagicMock(operationState=ThriftState.RUNNING_STATE, errorMessage=None),
            mock.MagicMock(operationState=ThriftState.RUNNING_STATE, errorMessage=None),
            mock.MagicMock(operationState=ThriftState.FINISHED_STATE, errorMessage=None),
        ]
        self.mock_cursor.poll.side_effect = poll_responses

        # Execute query
        wrapper.execute("SELECT 1")

        # Verify sleep was called twice (once for each PENDING state)
        self.assertEqual(mock_sleep.call_count, 2)
        mock_sleep.assert_called_with(5)

    @mock.patch("dbt.adapters.spark.connections.time.sleep")
    @mock.patch("dbt.adapters.spark.connections.time.time")
    def test_query_timeout_enforced(self, mock_time, mock_sleep):
        """Verify that query timeout is enforced when configured."""
        wrapper = PyhiveConnectionWrapper(self.mock_connection, poll_interval=5, query_timeout=10)
        wrapper._cursor = self.mock_cursor

        # Mock time progression: 0, 5, 11 (exceeds 10s timeout)
        mock_time.side_effect = [0, 5, 11]

        # Mock poll responses: always pending
        self.mock_cursor.poll.return_value = mock.MagicMock(
            operationState=ThriftState.RUNNING_STATE, errorMessage=None
        )

        # Execute query and expect timeout
        with self.assertRaises(DbtRuntimeError) as context:
            wrapper.execute("SELECT 1")

        self.assertIn("exceeded timeout", str(context.exception))
        self.assertIn("10 seconds", str(context.exception))

    @mock.patch("dbt.adapters.spark.connections.time.sleep")
    @mock.patch("dbt.adapters.spark.connections.time.time")
    def test_query_timeout_zero(self, mock_time, mock_sleep):
        """Verify that query_timeout of 0 causes immediate timeout."""
        wrapper = PyhiveConnectionWrapper(
            self.mock_connection, poll_interval=5, query_timeout=0, query_retries=0
        )
        wrapper._cursor = self.mock_cursor

        # Mock time progression: start_time=0, first check has elapsed > 0
        mock_time.side_effect = [0, 0.001]  # Even tiny elapsed time exceeds 0

        # Mock poll responses: query starts in pending state
        self.mock_cursor.poll.return_value = mock.MagicMock(
            operationState=ThriftState.RUNNING_STATE, errorMessage=None
        )

        # Execute query and expect immediate timeout
        with self.assertRaises(DbtRuntimeError) as context:
            wrapper.execute("SELECT 1")

        error_msg = str(context.exception)
        self.assertIn("exceeded timeout", error_msg.lower())
        self.assertIn("0 second", error_msg)

    @mock.patch("dbt.adapters.spark.connections.time.sleep")
    def test_connection_lost_during_polling_ttransport(self, mock_sleep):
        """Verify that TTransportException during polling is handled gracefully."""
        wrapper = PyhiveConnectionWrapper(self.mock_connection, poll_interval=5, query_retries=0)
        wrapper._cursor = self.mock_cursor

        # Mock poll responses: pending -> connection lost
        from thrift.transport.TTransport import TTransportException

        poll_responses = [
            mock.MagicMock(operationState=ThriftState.RUNNING_STATE, errorMessage=None),
            TTransportException(message="TSocket read 0 bytes"),
        ]
        self.mock_cursor.poll.side_effect = poll_responses

        # Execute query and expect helpful error
        with self.assertRaises(DbtRuntimeError) as context:
            wrapper.execute("SELECT 1")

        error_msg = str(context.exception).lower()
        self.assertIn("connection lost", error_msg)
        self.assertIn("long-running queries", error_msg)
        # Verify original exception type is included
        self.assertIn("ttransportexception", error_msg)

    @mock.patch("dbt.adapters.spark.connections.time.sleep")
    def test_connection_reset_during_polling(self, mock_sleep):
        """Verify that ConnectionResetError during polling is handled."""
        wrapper = PyhiveConnectionWrapper(self.mock_connection, poll_interval=5, query_retries=0)
        wrapper._cursor = self.mock_cursor

        # Mock poll responses: pending -> connection reset
        poll_responses = [
            mock.MagicMock(operationState=ThriftState.RUNNING_STATE, errorMessage=None),
            ConnectionResetError("Connection reset by peer"),
        ]
        self.mock_cursor.poll.side_effect = poll_responses

        # Execute query and expect helpful error
        with self.assertRaises(DbtRuntimeError) as context:
            wrapper.execute("SELECT 1")

        error_msg = str(context.exception).lower()
        self.assertIn("connection lost", error_msg)
        self.assertIn("connectionreseterror", error_msg)

    @mock.patch("dbt.adapters.spark.connections.time.sleep")
    def test_broken_pipe_during_polling(self, mock_sleep):
        """Verify that BrokenPipeError during polling is handled."""
        wrapper = PyhiveConnectionWrapper(self.mock_connection, poll_interval=5, query_retries=0)
        wrapper._cursor = self.mock_cursor

        # Mock poll responses: pending -> broken pipe
        poll_responses = [
            mock.MagicMock(operationState=ThriftState.RUNNING_STATE, errorMessage=None),
            BrokenPipeError("Broken pipe"),
        ]
        self.mock_cursor.poll.side_effect = poll_responses

        # Execute query and expect helpful error
        with self.assertRaises(DbtRuntimeError) as context:
            wrapper.execute("SELECT 1")

        error_msg = str(context.exception).lower()
        self.assertIn("connection lost", error_msg)
        self.assertIn("brokenpipeerror", error_msg)

    @mock.patch("dbt.adapters.spark.connections.time.sleep")
    def test_eof_error_during_polling(self, mock_sleep):
        """Verify that EOFError during polling is handled."""
        wrapper = PyhiveConnectionWrapper(self.mock_connection, poll_interval=5, query_retries=0)
        wrapper._cursor = self.mock_cursor

        # Mock poll responses: pending -> EOF
        poll_responses = [
            mock.MagicMock(operationState=ThriftState.RUNNING_STATE, errorMessage=None),
            EOFError("EOF when reading a line"),
        ]
        self.mock_cursor.poll.side_effect = poll_responses

        # Execute query and expect helpful error
        with self.assertRaises(DbtRuntimeError) as context:
            wrapper.execute("SELECT 1")

        error_msg = str(context.exception).lower()
        self.assertIn("connection lost", error_msg)
        self.assertIn("eoferror", error_msg)

    @mock.patch("dbt.adapters.spark.connections.time.sleep")
    def test_no_timeout_by_default(self, mock_sleep):
        """Verify that queries can run indefinitely if no timeout is set."""
        wrapper = PyhiveConnectionWrapper(
            self.mock_connection, poll_interval=5, query_timeout=None
        )
        wrapper._cursor = self.mock_cursor

        # Mock poll responses: many pending states, then success
        poll_responses = [
            mock.MagicMock(operationState=ThriftState.RUNNING_STATE, errorMessage=None)
            for _ in range(100)
        ] + [mock.MagicMock(operationState=ThriftState.FINISHED_STATE, errorMessage=None)]
        self.mock_cursor.poll.side_effect = poll_responses

        # Execute query - should not timeout
        wrapper.execute("SELECT 1")

        # Verify it polled many times
        self.assertEqual(mock_sleep.call_count, 100)

    @mock.patch("dbt.adapters.spark.connections.time.sleep")
    def test_error_message_from_server(self, mock_sleep):
        """Verify that error messages from the server are properly raised."""
        wrapper = PyhiveConnectionWrapper(self.mock_connection, poll_interval=5)
        wrapper._cursor = self.mock_cursor

        # Mock poll response with error message
        error_response = mock.MagicMock(
            operationState=ThriftState.ERROR_STATE, errorMessage="Table not found: test_table"
        )
        self.mock_cursor.poll.return_value = error_response

        # Execute query and expect database error
        with self.assertRaises(DbtDatabaseError) as context:
            wrapper.execute("SELECT * FROM test_table")

        self.assertIn("Table not found", str(context.exception))

    @mock.patch("dbt.adapters.spark.connections.time.sleep")
    def test_cancelled_query_raises_error(self, mock_sleep):
        """Verify that cancelled queries raise an appropriate error."""
        wrapper = PyhiveConnectionWrapper(self.mock_connection, poll_interval=5)
        wrapper._cursor = self.mock_cursor

        # Mock poll response: cancelled state without error message
        cancelled_response = mock.MagicMock(
            operationState=ThriftState.CANCELED_STATE, errorMessage=None
        )
        self.mock_cursor.poll.return_value = cancelled_response

        # Execute query and expect error
        with self.assertRaises(DbtDatabaseError) as context:
            wrapper.execute("SELECT 1")

        self.assertIn("CANCELED", str(context.exception))

    @mock.patch("dbt.adapters.spark.connections.time.sleep")
    def test_custom_poll_interval(self, mock_sleep):
        """Verify that custom poll interval is respected."""
        wrapper = PyhiveConnectionWrapper(self.mock_connection, poll_interval=10)
        wrapper._cursor = self.mock_cursor

        # Mock poll responses: pending -> success
        poll_responses = [
            mock.MagicMock(operationState=ThriftState.RUNNING_STATE, errorMessage=None),
            mock.MagicMock(operationState=ThriftState.FINISHED_STATE, errorMessage=None),
        ]
        self.mock_cursor.poll.side_effect = poll_responses

        # Execute query
        wrapper.execute("SELECT 1")

        # Verify sleep was called with custom interval
        mock_sleep.assert_called_once_with(10)

    @mock.patch("dbt.adapters.spark.connections.time.sleep")
    def test_non_connection_exception_reraised(self, mock_sleep):
        """Verify that non-connection exceptions are re-raised as-is."""
        wrapper = PyhiveConnectionWrapper(self.mock_connection, poll_interval=5, query_retries=0)
        wrapper._cursor = self.mock_cursor

        # Mock poll responses: pending -> some other exception
        poll_responses = [
            mock.MagicMock(operationState=ThriftState.RUNNING_STATE, errorMessage=None),
            ValueError("Some unexpected error"),
        ]
        self.mock_cursor.poll.side_effect = poll_responses

        # Execute query and expect ValueError to be re-raised (not wrapped)
        with self.assertRaises(ValueError) as context:
            wrapper.execute("SELECT 1")

        self.assertIn("unexpected error", str(context.exception))

    @mock.patch("dbt.adapters.spark.connections.time.sleep")
    def test_exception_caught_by_type_not_message(self, mock_sleep):
        """Verify that exceptions are caught by type, not by parsing message strings."""
        wrapper = PyhiveConnectionWrapper(self.mock_connection, poll_interval=5, query_retries=0)
        wrapper._cursor = self.mock_cursor

        # Create a ConnectionResetError with a message that doesn't contain typical keywords
        # If we were parsing strings, this might not be caught, but it should be caught by type
        unusual_error = ConnectionResetError("xyz123")  # No "reset" or "peer" in message

        poll_responses = [
            mock.MagicMock(operationState=ThriftState.RUNNING_STATE, errorMessage=None),
            unusual_error,
        ]
        self.mock_cursor.poll.side_effect = poll_responses

        # Execute query and expect it to be caught as connection error
        with self.assertRaises(DbtRuntimeError) as context:
            wrapper.execute("SELECT 1")

        error_msg = str(context.exception)
        # Should be wrapped as connection lost error (caught by type)
        self.assertIn("connection lost", error_msg.lower())
        # Should include the exception type name
        self.assertIn("connectionreseterror", error_msg.lower())
        # Should include original message even though it's unusual
        self.assertIn("xyz123", error_msg)

    @mock.patch("dbt.adapters.spark.connections.time.sleep")
    def test_chained_exception_preserved(self, mock_sleep):
        """Verify that exception chaining is preserved for connection errors."""
        wrapper = PyhiveConnectionWrapper(self.mock_connection, poll_interval=5, query_retries=0)
        wrapper._cursor = self.mock_cursor

        # Mock poll responses: pending -> connection lost
        original_error = ConnectionResetError("Connection reset by peer")
        poll_responses = [
            mock.MagicMock(operationState=ThriftState.RUNNING_STATE, errorMessage=None),
            original_error,
        ]
        self.mock_cursor.poll.side_effect = poll_responses

        # Execute query and verify exception chaining
        with self.assertRaises(DbtRuntimeError) as context:
            wrapper.execute("SELECT 1")

        # Verify the original exception is preserved in the chain
        self.assertIsNotNone(context.exception.__cause__)
        self.assertIsInstance(context.exception.__cause__, ConnectionResetError)

    @mock.patch("dbt.adapters.spark.connections.time.sleep")
    def test_query_retry_on_connection_loss(self, mock_sleep):
        """Verify that queries are retried on connection loss."""
        wrapper = PyhiveConnectionWrapper(self.mock_connection, poll_interval=5, query_retries=2)
        wrapper._cursor = self.mock_cursor

        # Mock cursor() to return fresh cursors on retry
        fresh_cursor = mock.MagicMock()
        self.mock_connection.cursor.return_value = fresh_cursor

        # First attempt: pending -> connection lost
        # Second attempt: pending -> success
        first_attempt = [
            mock.MagicMock(operationState=ThriftState.RUNNING_STATE, errorMessage=None),
            ConnectionResetError("Connection reset by peer"),
        ]
        second_attempt = [
            mock.MagicMock(operationState=ThriftState.RUNNING_STATE, errorMessage=None),
            mock.MagicMock(operationState=ThriftState.FINISHED_STATE, errorMessage=None),
        ]

        self.mock_cursor.poll.side_effect = first_attempt
        fresh_cursor.poll.side_effect = second_attempt

        # Execute query - should succeed after retry
        wrapper.execute("SELECT 1")

        # Verify sleep was called for polling + retry delay
        self.assertGreater(mock_sleep.call_count, 1)

    @mock.patch("dbt.adapters.spark.connections.time.sleep")
    def test_query_retry_exhausted(self, mock_sleep):
        """Verify that error is raised when all retries are exhausted."""
        wrapper = PyhiveConnectionWrapper(self.mock_connection, poll_interval=5, query_retries=1)
        wrapper._cursor = self.mock_cursor

        # Mock cursor() to return fresh cursors
        fresh_cursor1 = mock.MagicMock()
        fresh_cursor2 = mock.MagicMock()
        self.mock_connection.cursor.side_effect = [fresh_cursor1, fresh_cursor2]

        # Both attempts fail with connection loss
        connection_error = ConnectionResetError("Connection reset by peer")
        self.mock_cursor.poll.side_effect = [
            mock.MagicMock(operationState=ThriftState.RUNNING_STATE, errorMessage=None),
            connection_error,
        ]
        fresh_cursor1.poll.side_effect = [
            mock.MagicMock(operationState=ThriftState.RUNNING_STATE, errorMessage=None),
            connection_error,
        ]

        # Execute query and expect failure after retries
        with self.assertRaises(DbtRuntimeError) as context:
            wrapper.execute("SELECT 1")

        error_msg = str(context.exception)
        self.assertIn("after 2 attempts", error_msg)
        self.assertIn("consider increasing 'query_retries'", error_msg.lower())

    @mock.patch("dbt.adapters.spark.connections.time.sleep")
    def test_no_retry_on_database_errors(self, mock_sleep):
        """Verify that database errors are not retried, only connection errors."""
        wrapper = PyhiveConnectionWrapper(self.mock_connection, poll_interval=5, query_retries=2)
        wrapper._cursor = self.mock_cursor

        # Mock poll response with database error (not connection error)
        error_response = mock.MagicMock(
            operationState=ThriftState.ERROR_STATE, errorMessage="Table not found: test_table"
        )
        self.mock_cursor.poll.return_value = error_response

        # Execute query - should fail immediately without retry
        with self.assertRaises(DbtDatabaseError) as context:
            wrapper.execute("SELECT * FROM test_table")

        # Should not retry - only 1 attempt
        self.assertEqual(self.mock_cursor.execute.call_count, 1)


if __name__ == "__main__":
    unittest.main()
